package com.sillycycle.bagleyd.pyraminx;

/*
 * @(#)PyraminxFrame.java
 *
 * Copyright 1994 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Label;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ArgumentParser;
import com.sillycycle.bagleyd.util.Colour;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComponentUtil;
import com.sillycycle.bagleyd.util.Icon;
import com.sillycycle.bagleyd.util.IODialog;
import com.sillycycle.bagleyd.util.OKDialog;
import com.sillycycle.bagleyd.util.Spinner;

/**
 * The <code>PyraminxFrame</code> class is to be embedded in a Web page
 * or run as an application.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class PyraminxFrame extends JFrame implements ActionListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_FACETS = 7;
	static final int MIN_SPEED = 1;
	static final int MAX_SPEED = 50;
	static final int NEVER = -1;
	static final String SCOREFILE = "pyraminx.scores";
	static final String[] COMMAND_STRINGS = new String[] {
		"Get Buffer", "Write Buffer", "Show Buffer",
		"Undo", "Redo", "Clear",
		"Practice", "Randomize", "Auto-solve",
		"Increment", "Decrement", "Orientize", "Sticky",
		"Speed >", "Slow <", "Sound @",
		"Description ?", "Features !", "References ^", "About",
		"Get File", "Write File", "Exit"};
	static final char[] COMMAND_CHARS = new char[] {
		'g', 'w', '*',
		'u', 'r', 'c', 'p', 'z', 's',
		'i', 'd', 'o', 'y', '>', '<', '@',
		'?', '!', '^', 'a', 'G', 'W', 'x'};

	static ArgumentParser argumentParser = null;
	Icon mediumIcon = null;
	Icon largeIcon = null;
	int widthValue = 420;
	int heightValue = 210;
	static int xValue = 0;
	static int yValue = 0;
	String monoString = "BRYG";
	Color[] faceColor = new Color[] {
		Color.blue, Color.red, Color.yellow, Color.green
	};
	Color borderColor = Color.black;
	Color fgColor = Color.cyan;
	Color bgColor = Colour.steelBlue;
	boolean controlValue = true;
	boolean sound = false;
	int sizeValue = PyraminxInterface.DEFAULTFACETS;
	int oldFacet = sizeValue;
	int modeValue = PyraminxInterface.DEFAULT_MODE;
	boolean orientValue = PyraminxInterface.DEFAULT_ORIENT;
	boolean stickyValue = PyraminxInterface.DEFAULTSTICKY;
	boolean practiceValue = PyraminxInterface.DEFAULT_PRACTICE;
	int delayValue = 10;
	int speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
	boolean debug = false;
	JButton button = null;
	protected JPopupMenu popup = null;
	JPanel topPanel = new JPanel();
	JComboBox<Object> modeComboBox = new JComboBox<>();
	JCheckBox orientCheckBox = new JCheckBox("Oriented");
	JCheckBox stickyCheckBox = new JCheckBox("Sticky");
	JCheckBox practiceCheckBox = new JCheckBox("Practice");
	Pyraminx puzzle;
	int movesDisplay = 0;
	String recordDisplay = "NEVER";
	String messageDisplay = "Play Pyraminx! (use mouse or keypad)";
	JTextField movesValueLabel = new JTextField(Integer.toString(
		movesDisplay));
	JTextField recordValueLabel = new JTextField(recordDisplay);
	Label messageLabel = new Label(messageDisplay, Label.CENTER);
	// JLabel constantly resizes
	OKDialog aDialog = null;
	StringBuilder stringBuilderSave;
	IODialog ioDialog = null;
	int[][][] puzzleRecord = null;
	JMenuBar menuBar;
	JMenuItem detachMenuItem;
	JPanel mainPanel = new JPanel();
	JPanel secondaryPanel = new JPanel();
	JFrame frame;
	Spinner facetChanger = null, speedChanger = null;
	//Slider facetChanger = null, speedChanger = null;

	public String writePuzzleString() {
		return this.stringBuilderSave.toString();
	}

	public void getPuzzleString(String s) {
		puzzle.stringSave = s;
	}

	public void processKey(char key) {
		puzzle.processKey(key);
	}

	public boolean getToggleSound() {
		return sound;
	}

	public void toggleSound() {
		sound = !sound;
	}

	public String getOpt(String string) {
		return argumentParser.getOption(string);
	}

	public int getOpt(String param, int value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Integer.parseInt(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public boolean getOpt(String param, boolean value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			if (string.equalsIgnoreCase("true") ||
					string.equalsIgnoreCase("yes") ||
					string.equalsIgnoreCase("on"))
				return true;
			else if (string.equalsIgnoreCase("false") ||
					string.equalsIgnoreCase("no") ||
					string.equalsIgnoreCase("off"))
				return false;
			return (Integer.parseInt(string) != 0);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public Color getOpt(String param, Color value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Color.decode(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public void getOpts() {
		String string;

		widthValue = getOpt("windowWidth", widthValue);
		heightValue = getOpt("windowHeight", heightValue);
		xValue = getOpt("windowX", xValue);
		yValue = getOpt("windowY", yValue);
		for (int i = 0; i < faceColor.length; i++)
			faceColor[i] = getOpt("faceColor" + i, faceColor[i]);
		borderColor = getOpt("borderColor", borderColor);
		fgColor = getOpt("fgColor", fgColor);
		bgColor = getOpt("bgColor", bgColor);
		sound = getOpt("sound", sound);
		controlValue = getOpt("control", controlValue);
		sizeValue = getOpt("size", sizeValue);
		if (sizeValue < PyraminxInterface.MIN_FACETS)
			sizeValue = PyraminxInterface.DEFAULTFACETS;
		modeValue = getOpt("mode", modeValue);
		string = getOpt("both");
		if (string != null)
			modeValue = PyraminxInterface.BOTH;
		if (modeValue < PyraminxInterface.PERIOD2 ||
				modeValue > PyraminxInterface.BOTH)
			modeValue = PyraminxInterface.DEFAULT_MODE;
		orientValue = getOpt("orient", orientValue);
		stickyValue = getOpt("sticky", stickyValue);
		practiceValue = getOpt("practice", practiceValue);
		delayValue = getOpt("delay", delayValue);
	}

	public void showBuffer() {
		if (ioDialog != null) {
			ioDialog.setVisible(true);
		}
	}

	public void showMessage(String msg) {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Pyraminx", msg, "OK",
				mediumIcon, false);
			aDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2, 365, 140);
			aDialog.setVisible(true);
		}
	}

	public void descriptionHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Description",
"Pyraminx - the original puzzle has 9 triangles per face" + NL +
"(size = 3) and has period 3 turning (i.e. the face or points" + NL +
"turn in 120 degree intervals).  The puzzle was designed by Uwe" + NL +
"Meffert and called the Pyraminx.  This has 2^5*3^8*6!/2 or" + NL +
"75,582,720 different comibinations." + NL +
"Another puzzle Senior Pyraminx 3x3x3 exists only on paper, it" + NL +
"has period 2 turning (i.e. edges turn with 180 degree" + NL +
"intervals) but the corners would fall off unless it had some" + NL +
"tricky mechanism.  (This may be the same as the Master Pyraminx" + NL +
"which has 446,965,972,992,000 different combinations)." + NL +
"Another puzzle (which was not widely distributed), the Junior" + NL +
"Pyraminx (and similarly the Junior Pyraminx Star, a octahedron" + NL +
"formed by two tetrahedra, this has 7!*3^6 or 3,674,160" + NL +
"different combinations), has 4 triangles (size = 2) per face." + NL +
"This puzzle has been recently reissued by Meffert as" + NL +
"Pyramorphix (http://www.mefferts-puzzles.com). At the time" + NL +
"I designed this computer puzzle thought that it had only period" + NL +
"2 turning (i.e the edges rotate).  It turns out the puzzle has" + NL +
"a period 4 turning (edges turn with 90 degree intervals) which" + NL +
"makes it analogous to the 2x2x2 Rubik's cube.  This puzzle" + NL +
"makes various non-tetrahedral shapes.  The puzzle contained" + NL +
"here has no period 4 turning flexibility." + NL +
"One is able to simulate Halpern's Tetrahedron or Pyraminx" + NL +
"Tetrahedron (period 3 turning and sticky mode).   Also one is" + NL +
"able to simulate one with variant turning (period 2 turning and" + NL +
"sticky mode)." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void featuresHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Features",
"Make sure you randomize the puzzle first before trying to move the" + NL +
"pieces for the first time." + NL +
" " + NL +
"Press \"mouse-left\" button to move a piece.  Release \"mouse-left\"" + NL +
"button on another piece on the same face and in the same row" + NL +
"(but not an adjacent piece or the move is ambiguous).  The pieces" + NL +
"will then turn towards where the mouse button was released." + NL +
"on the same period is ambiguous and the puzzle will not turn." + NL +
" " + NL +
"Click \"mouse-right\" button, or press \"Z\" or \"z\" keys, to" + NL +
"randomize the puzzle (this must be done first to set a new record)." + NL +
"One must double click on \"mouse-right\" if the puzzle is currently" + NL +
"being worked on." + NL +
" " + NL +
"Press \"g\" key to get (read) a saved puzzle to screen." + NL +
" " + NL +
"Press \"w\" key to save (write) a puzzle to screen." + NL +
" " + NL +
"Press \"G\" key to get (read) a saved puzzle to file." + NL +
" " + NL +
"Press \"W\" key to save (write) a puzzle to file." + NL +
" " + NL +
"Press \"U\" or \"u\" keys to undo a move." + NL +
" " + NL +
"Press \"R\" or \"r\" keys to redo a move." + NL +
" " + NL +
"Press \"C\" or \"c\" keys to clear the puzzle." + NL +
" " + NL +
"Press \"P\" or \"p\" keys, to toggle the practice mode (in practice" + NL +
"mode the record should say \"practice\").  This is good for" + NL +
"learning moves and experimenting." + NL +
" " + NL +
"Press \"S\" or \"s\" keys to start auto-solver.  Only works where" + NL +
"size < 4 in Period 2 mode (or sticky mode is set) and size < 5 in" + NL +
"Period 3 mode." + NL +
" " + NL +
"Press \"O\" or \"o\" keys to toggle the orient mode.  (Default is off)." + NL +
"One has to orient the faces in orient mode, besides getting all" + NL +
"the faces to be the same color.  To do this one has to get the" + NL +
"lines to be oriented in the same direction, this only matters with" + NL +
"center pieces, if at all (i.e. those pieces not on a" + NL +
"corner or edge).  This does add complexity so there are 2 sets of" + NL +
"records." + NL +
" " + NL +
"Press \"2\", \"3\", \"B\", or \"b\" keys (not the keypad 2, 3)" + NL +
"to change modes to Period 2, Period 3, or Both." + NL +
" " + NL +
"Press \"Y\" or \"y\" to toggle \"Sticky\" mode (increase/decrease" + NL +
"is disabled here if \"Sticky\" mode is on)." + NL +
"\"Sticky\" and \"Period 2\" turning allows only the edges to turn, and" + NL +
"the 2 center rows turn together. It is as if the middle cut of the" + NL +
"three cuts did not exist." + NL +
"\"Sticky\" and \"Period 3\" turning allows only the faces to turn, it" + NL +
"is as if the middle cut of the three cuts did not exist." + NL +
"Beware, the \"Sticky\" mode is a hack and much could be done to improve" + NL +
"its look." + NL +
" " + NL +
"Press \"I\" or \"i\" keys to increase the number of pieces." + NL +
" " + NL +
"Press \"D\" or \"d\" keys to decrease the number of pieces." + NL +
" " + NL +
"Press \">\" or \".\" keys to speed up the movement of the pieces." + NL +
" " + NL +
"Press \"<\" or \",\" keys to slow up the movement of the pieces." + NL +
" " + NL +
"Press \"@\" key to toggle the sound." + NL +
" " + NL +
"Press \"Esc\" key to hide program." + NL +
" " + NL +
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill program if an application." + NL +
" " + NL +
"Use the key pad or arrow keys to move without the mouse." + NL +
"Key pad (better with Num Lock on) is defined for Pyraminx as:" + NL +
"  /    Counterclockwise" + NL +
" " + NL +
"   8 9  Up, Upper Right" + NL +
"   ^" + NL +
"4<5>6  Left, Clockwise, Right" + NL +
"   v" + NL +
"1 2     Lower Left, Down" + NL +
" " + NL +
"Use the shift keys to access \"Period 3\" turns from \"Both\" mode," + NL +
"otherwise it assumes \"Period 2\" turning.  Faces and points turn in" + NL +
"\"Period 3\" and edges (2 points) turn in \"Period 2\"." + NL +
"Use the control key and the left mouse button, keypad, or arrow" + NL +
"keys to move the whole puzzle.  This is not recorded as a turn." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void referencesHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"References",
"James G. Nourse, The Simple Solutions to Cubic Puzzles, " +
"Bantam Books, New York, 1981. pp 8-15." + NL +
"Tom Werneck, Mastering the Magic Pyramid, Evans Brothers " +
"Limited, London, 1981. pp 109-111." + NL +
"Douglas R. Hofstadter, Beyond Rubik's Cube: spheres, pyramids, " +
"dodecahedrons and God knows what else, Scientific American," + NL +
"July 1982, pp 16-31." + NL +
"John Ewing & Czes Kosniowski, Puzzle it Out: Cubes Groups and " +
"Puzzles, Cambridge University Press, New York, 1982, pp 60-61." + NL +
"Magic Cubes 1996 Catalog of Dr. Christoph Bandelow." + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void aboutHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"About",
"Java PyraminxFrame Version 8.7.2" + NL +
"Send bugs (reports or fixes) to the author:" +
"David Bagley <bagleyd AT verizon.net>" + NL +
"The latest version is at: " +
"https://www.sillycycle.com/puzzles.html" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public Icon getIcon(String string, int sx, int sy) {
		URL url = PyraminxFrame.class.getResource("/" + string);

		if (url == null)
			return null;
		return new Icon(Toolkit.getDefaultToolkit().
			getImage(url), sx, sy);
	}

	// TODO Make more reliable, also should cache
	public void playClip(String string) {
		URL url = PyraminxFrame.class.getResource("/" + string);

		if (url == null)
			return;
		try {
			AudioInputStream audioInputStream =
				AudioSystem.getAudioInputStream(url);
			DataLine.Info info = new DataLine.Info(Clip.class,
				audioInputStream.getFormat());
			Clip clip = (Clip)AudioSystem.getLine(info);
			clip.open(audioInputStream);
			clip.start();
		} catch (IOException | LineUnavailableException | UnsupportedAudioFileException e) {
			e.printStackTrace();
		}
	}

	public void playMoveAudio() {
		playClip(PyraminxInterface.SOUNDS[PyraminxInterface.Sounds.MOVE.ordinal()]);
	}

	@Override
	public void setSize(int width, int height) {
		if (puzzle == null)
			return;
		puzzle.setBounds(0, 0, width, height);
		/*repaint();*/ // causes flashing
	}

	@Override
	public void paint(Graphics g) {
		setSize(secondaryPanel.getSize().width,
			secondaryPanel.getSize().height - topPanel.getSize().height);
		super.paint(g);
	}

	public String readUserScore() {
		StringBuilder scoreString = new StringBuilder();

		for (int i = 0; i < puzzleRecord.length; i++) {
			for (int j = 0; j < puzzleRecord[i].length; j++) {
				for (int k = 0; k < puzzleRecord[i][j].length; k++) {
					scoreString.append(puzzleRecord[i][j][k]);
					if (k < puzzleRecord[i][j].length - 1)
						scoreString.append(" ");
				}
				if (j < puzzleRecord[i].length - 1
						|| i < puzzleRecord.length - 1)
					scoreString.append(NL);
			}
			if (i < puzzleRecord.length - 1)
				scoreString.append(NL);
		}
		return scoreString.toString();
	}

	public void writeScore() {
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				SCOREFILE, false))) {
			pr.println(readUserScore());
		} catch (Exception e) {
			//e.printStackTrace();
		}
	}

	public void initRecords(String cookieValue, boolean store) {
		String[] st = null;
		int field = 0;

		if (cookieValue != null)
			st = cookieValue.split("[+/\\s]+");
		if (debug)
			System.out.println("initRecords: " + cookieValue);
		puzzleRecord = new int[PyraminxInterface.MAX_MODES][2]
			[MAX_FACETS - PyraminxInterface.MIN_FACETS + 2];
		for (int i = 0; i < puzzleRecord.length; i++) {
		  for (int j = 0; j < puzzleRecord[i].length; j++) {
		    for (int k = 0; k < puzzleRecord[i][j].length; k++) {
			if (st == null || st.length <= field) {
				//build it from the array
				puzzleRecord[i][j][k] =
					PyraminxScores.PUZZLE_SCORES[i][j][k];
			} else {
				puzzleRecord[i][j][k] =
					Integer.parseInt(st[field++]);
			}
		    }
		  }
		}
		if (store && st == null)
			writeScore(); // if no cookie, then make one...
	}

	public void readScoreRecords() {
		try (BufferedReader br = new BufferedReader(
				new FileReader(SCOREFILE))) {
			StringBuilder scoreData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null) {
				scoreData.append(lineData);
				scoreData.append(NL);
			}
			initRecords(scoreData.toString(), true);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot open " + SCOREFILE);
			initRecords(null, true);
		}
	}

	public String readTextArea() {
		return ioDialog.getText();
	}

	public void writeTextArea(String s) {
		ioDialog.setText(s);
	}

	public void shuffleDown() {
		ComponentUtil.findJFrame(this).toBack();
	}

	public void quit() {
		System.exit(0);
	}

	void printRecord(int size, int mode,
			boolean orient, boolean sticky,
			boolean practice) {
		int i = mode - PyraminxInterface.PERIOD2;
		int j = (orient) ? 1 : 0;
		int k = (sticky) ? MAX_FACETS -
			PyraminxInterface.MIN_FACETS + 1 :
			size - PyraminxInterface.MIN_FACETS;

		if (puzzleRecord == null)
			return;
		if (practice) {
			recordValueLabel.setText("PRACTICE");
		} else if (!sticky && size > MAX_FACETS) {
			recordValueLabel.setText("NONE");
		} else if (puzzleRecord[i][j][k] <= NEVER) {
			recordValueLabel.setText("NEVER");
		} else {
			recordValueLabel.setText(Integer.toString(
				puzzleRecord[i][j][k]));
		}
	}

	boolean handleSolved(int counter, int size, int mode,
			boolean orient, boolean sticky,
			boolean practice) {
		int i = mode - PyraminxInterface.PERIOD2;
		int j = (orient) ? 1 : 0;
		int k = (sticky) ? MAX_FACETS -
			PyraminxInterface.MIN_FACETS + 1 :
			size - PyraminxInterface.MIN_FACETS;
		int notJ = (orient) ? 0 : 1;

		if (puzzleRecord != null &&
				(sticky || size <= MAX_FACETS) &&
				(counter < puzzleRecord[i][j][k] ||
				puzzleRecord[i][j][k] <= NEVER)) {
			readScoreRecords(); /* Maybe its been updated by another */
			puzzleRecord[i][j][k] = counter;
			if (size < 4 ||
			    mode == PyraminxInterface.PERIOD2 ||
			    (orient &&
			    (counter < puzzleRecord[i][notJ][k] ||
			    puzzleRecord[i][notJ][k] <= NEVER))) {
				puzzleRecord[i][notJ][k] = counter;
			}
			writeScore();
			printRecord(size, mode,
				orient, sticky, practice);
			return true;
		}
		return false;
	}

	public void initScore() {
		if (puzzleRecord == null) {
			readScoreRecords();
			printRecord(sizeValue, modeValue,
				orientValue, stickyValue, practiceValue);
		}
	}

	class FacetChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = sizeValue;

			if (stickyValue) {
				facetChanger.setValue(oldValue);
				return;
			}
			oldFacet = sizeValue = facetChanger.getValue();
			if (oldValue != sizeValue) {
				puzzle.setPuzzleSize(sizeValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(sizeValue, modeValue,
					orientValue, stickyValue,
					practiceValue);
			}
		}
	}

	class SpeedChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = speedValue;

			speedValue = speedChanger.getValue();
			if (oldValue != speedValue) {
				puzzle.setDelay(MAX_SPEED + MIN_SPEED - 1 -
					speedValue);
			}
		}
	}

	class ComboBoxActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int modeIndex = modeComboBox.getSelectedIndex()
					+ PyraminxInterface.PERIOD2;

			if (modeIndex != -1 && modeValue != modeIndex) {
				modeValue = modeIndex;
				puzzle.setPuzzlePeriod(modeValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(sizeValue, modeValue,
					orientValue, stickyValue,
					practiceValue);
			}
		}
	}

	class OrientItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = orientValue;

			orientValue = orientCheckBox.isSelected();
			if (oldValue != orientValue) {
				puzzle.setPuzzleOrient(orientValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(sizeValue, modeValue,
					orientValue, stickyValue,
					practiceValue);
			}
		}
	}

	class StickyItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = stickyValue;

			stickyValue = stickyCheckBox.isSelected();
			if (oldValue != stickyValue) {
				if ((stickyValue && sizeValue != 4) ||
				    (!stickyValue && oldFacet != 4)) {
					if (stickyValue)
						sizeValue = 4;
					else
						sizeValue = oldFacet;
					facetChanger.setValue(sizeValue);
					puzzle.setPuzzleSize(sizeValue);
				}
				puzzle.setPuzzleSticky(stickyValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(sizeValue, modeValue,
					orientValue, stickyValue,
					practiceValue);
			}
		}
	}

	class PracticeItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = practiceValue;

			practiceValue = practiceCheckBox.isSelected();
			if (oldValue != practiceValue) {
				puzzle.setPuzzlePractice(practiceValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				if (!practiceValue)
					messageLabel.setText(
						"Randomize to start");
				printRecord(sizeValue, modeValue,
					orientValue, stickyValue,
					practiceValue);
			}
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		String command = event.getActionCommand();

		for (int opt = 0; opt < COMMAND_STRINGS.length; opt++) {
			if (command.equals(COMMAND_STRINGS[opt])) {
				processKey(COMMAND_CHARS[opt]);
				return;
			}
		}
	}

	public void callback(int reason) {
		messageDisplay = "                                                                                ";
		switch(reason) {
		case PyraminxInterface.ACTION_RESTORE:
			if (practiceValue)
				recordValueLabel.setText("PRACTICE");
			movesDisplay = 0;
			break;
		case PyraminxInterface.ACTION_RESET:
			movesDisplay = 0;
			break;
		case PyraminxInterface.ACTION_AMBIGUOUS:
			messageDisplay = "Ambiguous move";
			break;
		case PyraminxInterface.ACTION_ILLEGAL:
			if (practiceValue || puzzle.checkStarted())
				messageDisplay = "Illegal move";
			else
				messageDisplay = "Randomize to start";
			break;
		case PyraminxInterface.ACTION_MOVED:
			movesDisplay++;
			break;
		case PyraminxInterface.ACTION_CONTROL:
			break;
		case PyraminxInterface.ACTION_SOLVED:
			if (practiceValue)
				movesDisplay = 0;
			else {
				if (puzzle.getCheat())
					messageDisplay = "No cheating!!"; //username
				else if (handleSolved(movesDisplay, sizeValue,
						modeValue, orientValue,
						stickyValue, practiceValue))
					messageDisplay = "Best solution, congratulations!!"; //username
				else
					messageDisplay = "Solved!";
			}
			//showMessage(messageDisplay);
			break;
		case PyraminxInterface.ACTION_PRACTICE:
			movesDisplay = 0;
			practiceValue = !practiceValue;
			if (!practiceValue)
				messageDisplay = "Randomize to start";
			printRecord(sizeValue, modeValue,
				orientValue, stickyValue, practiceValue);
			puzzle.setPuzzlePractice(practiceValue);
			practiceCheckBox.setSelected(practiceValue);
			break;
		case PyraminxInterface.ACTION_RANDOMIZE:
			movesDisplay = 0;
			practiceValue = false;
			puzzle.setPuzzlePractice(practiceValue);
			break;
		case PyraminxInterface.ACTION_INCREMENT:
			if (!stickyValue) {
				movesDisplay = 0;
				sizeValue++;
				oldFacet = sizeValue;
				printRecord(sizeValue, modeValue,
					orientValue, stickyValue,
					practiceValue);
				puzzle.setPuzzleSize(sizeValue);
				if (facetChanger != null)
					facetChanger.setValue(sizeValue);
			}
			break;
		case PyraminxInterface.ACTION_DECREMENT:
			if (!stickyValue) {
				movesDisplay = 0;
				sizeValue--;
				oldFacet = sizeValue;
				printRecord(sizeValue, modeValue,
					orientValue, stickyValue,
					practiceValue);
				puzzle.setPuzzleSize(sizeValue);
				if (facetChanger != null)
					facetChanger.setValue(sizeValue);
			}
			break;
		case PyraminxInterface.ACTION_ORIENTIZE:
			movesDisplay = 0;
			orientValue = !orientValue;
			printRecord(sizeValue, modeValue,
				orientValue, stickyValue, practiceValue);
			puzzle.setPuzzleOrient(orientValue);
			orientCheckBox.setSelected(orientValue);
			break;
		case PyraminxInterface.ACTION_PERIOD2:
		case PyraminxInterface.ACTION_PERIOD3:
		case PyraminxInterface.ACTION_BOTH:
			movesDisplay = 0;
			modeValue = reason -
				PyraminxInterface.ACTION_PERIOD2 +
				PyraminxInterface.PERIOD2;
			printRecord(sizeValue, modeValue,
				orientValue, stickyValue, practiceValue);
			puzzle.setPuzzlePeriod(modeValue);
			modeComboBox.setSelectedItem(
				PyraminxInterface.modeStrings[modeValue -
					PyraminxInterface.PERIOD2]);
			break;
		case PyraminxInterface.ACTION_STICKY:
			movesDisplay = 0;
			stickyValue = !stickyValue;
			if (stickyValue)
				sizeValue = 4;
			else
				sizeValue = oldFacet;
			printRecord(sizeValue, modeValue,
				orientValue, stickyValue, practiceValue);
			puzzle.setPuzzleSize(sizeValue);
			if (facetChanger != null)
				facetChanger.setValue(sizeValue);
			puzzle.setPuzzleSticky(stickyValue);
			stickyCheckBox.setSelected(stickyValue);
			break;
		case PyraminxInterface.ACTION_COMPUTED:
			break;
		case PyraminxInterface.ACTION_UNDO:
			movesDisplay--;
			break;
		case PyraminxInterface.ACTION_REDO:
			movesDisplay++;
			break;
		case PyraminxInterface.ACTION_SPEED:
			{
				int oldDelay, delay;

				delay = puzzle.getDelay();
				oldDelay = delay;
				if (delay > MAX_SPEED - MIN_SPEED)
					delay = MAX_SPEED - MIN_SPEED;
				if (delay < 0)
					delay = 0;
				if (delay != oldDelay) {
					puzzle.setDelay(delay);
				}
				if (speedChanger != null)
					speedChanger.setValue(MAX_SPEED +
						MIN_SPEED - 1 - delay);
			}
			return;
		default:
		}
		messageLabel.setText(messageDisplay);
		movesValueLabel.setText(Integer.toString(
			movesDisplay));
	}

	public PyraminxFrame(String title) {
		/*JPanel info1Panel = new JPanel();
		JPanel leftPanel = new JPanel();
		JPanel rightPanel = new JPanel();*/
		JPanel startPanel = new JPanel();
		JPanel info2Panel = new JPanel();
		JPanel movesPanel = new JPanel();
		JPanel recordPanel = new JPanel();
		JPanel checkBoxPanel = new JPanel();
		JPanel sliderPanel = new JPanel();
		JPanel comboBoxPanel = new JPanel();
		JPanel canvasPanel = new JPanel();
		int mode;

		try {
			mediumIcon = getIcon(PyraminxInterface.ICONS[PyraminxInterface.Icons.X32.ordinal()], 32, 32);
			largeIcon = getIcon(PyraminxInterface.ICONS[PyraminxInterface.Icons.X48.ordinal()], 48, 48);
		} catch (Exception e) {
			largeIcon = null;
		}
		ioDialog = new IODialog(ComponentUtil.findJFrame(this),
			PyraminxInterface.TITLE + ": IO", "Puzzle Position",
			"DONE");
		readScoreRecords();
		getOpts();
		speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBackground(Color.white);
		if (controlValue) {
			menuBar = new JMenuBar();
			JMenu menu;
			JMenuItem menuItem;

			JPopupMenu.setDefaultLightWeightPopupEnabled(false);
			menu = new JMenu("File");
			menuItem = new JMenuItem("Get Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_G);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_W);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Show Buffer");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Get File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Exit");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_X);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Play");
			menuItem = new JMenuItem("Undo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_U);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Redo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_R);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Clear");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_C);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Practice");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_P);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Randomize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_Z);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Auto-solve");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_S);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Increment");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_I);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Decrement");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_D);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Orientize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_O);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Sticky");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_Y);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Speed >");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Slow <");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Sound @");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Help");
			menuItem = new JMenuItem("Description ?");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Features !");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("References ^");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("About");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_A);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//setJMenuBar(menuBar); // needs to be in a panel
			mainPanel.add(menuBar, BorderLayout.PAGE_START);
			/*leftPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
				leftPanel.add(new JLabel("Move"));
			rightPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			rightPanel.add(new JLabel("Randomize"));
			leftPanel.add(getIcon("mouse-l.png", 16, 16));
			rightPanel.add(getIcon("mouse-r.png", 16, 16));
			info1Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 0));
			info1Panel.add(leftPanel);
			info1Panel.add(rightPanel);*/
			movesPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			movesPanel.add(new JLabel("Moves"));
			movesValueLabel.setColumns(5);
			movesValueLabel.setEditable(false);
			movesPanel.add(movesValueLabel);
			recordPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			recordPanel.add(new JLabel("Record"));
			recordValueLabel.setColumns(8);
			recordValueLabel.setEditable(false);
			recordPanel.add(recordValueLabel);
			info2Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 5));
			info2Panel.add(movesPanel);
			info2Panel.add(recordPanel);
			sliderPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			facetChanger = new Spinner("Size",
				sizeValue, PyraminxInterface.MIN_FACETS);
			/*facetChanger = new Slider("Size",
				sizeValue, PyraminxInterface.MIN_FACETS,
				((sizeValue > MAX_FACETS) ? sizeValue : MAX_FACETS));*/
			sliderPanel.add(facetChanger);
			for (mode = 0; mode < PyraminxInterface.MAX_MODES; mode++) {
				modeComboBox.addItem(
					PyraminxInterface.modeStrings[mode]);
			}
			modeComboBox.setSelectedItem(PyraminxInterface.
				modeStrings[modeValue - PyraminxInterface.PERIOD2]);
			comboBoxPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.CENTER));
			comboBoxPanel.add(modeComboBox);
			comboBoxPanel.add(new JLabel("Turning Mode"));
			sliderPanel.add(comboBoxPanel);
			speedChanger = new Spinner("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			/*speedChanger = new Slider("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			speedChanger.setWidth(150);*/
			checkBoxPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			checkBoxPanel.add(orientCheckBox);
			orientCheckBox.setSelected(orientValue);
			checkBoxPanel.add(stickyCheckBox);
			stickyCheckBox.setSelected(stickyValue);
			checkBoxPanel.add(practiceCheckBox);
			practiceCheckBox.setSelected(practiceValue);
			if (practiceValue)
				recordValueLabel.setText("PRACTICE");
			topPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.CENTER));
			/*topPanel.add(info1Panel);*/
			topPanel.add(info2Panel);
			topPanel.add(sliderPanel);
			topPanel.add(speedChanger);
			topPanel.add(checkBoxPanel);
			topPanel.add(messageLabel);
			startPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			startPanel.add(topPanel);
			secondaryPanel.setLayout(new BorderLayout());
			secondaryPanel.add(startPanel, BorderLayout.PAGE_START);
		}
		if (stickyValue)
			sizeValue = 4;
		puzzle = new Pyraminx(this,
			sizeValue, modeValue,
			orientValue, stickyValue, practiceValue,
			false, false, monoString,
			faceColor[0], faceColor[1], faceColor[2],
			faceColor[3],
			borderColor, fgColor, bgColor,
			MAX_SPEED + MIN_SPEED - 1 - speedValue);
		canvasPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		canvasPanel.add(puzzle);
		secondaryPanel.add(canvasPanel, BorderLayout.CENTER);
		mainPanel.add(secondaryPanel, BorderLayout.CENTER);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		if (controlValue) {
			printRecord(sizeValue, modeValue,
				orientValue, stickyValue, practiceValue);
			facetChanger.addChangeListener(new FacetChangeListener());
			modeComboBox.addActionListener(new ComboBoxActionListener());
			orientCheckBox.addItemListener(new OrientItemListener());
			stickyCheckBox.addItemListener(new StickyItemListener());
			practiceCheckBox.addItemListener(new PracticeItemListener());
			speedChanger.addChangeListener(new SpeedChangeListener());
		}
		canvasPanel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseEntered(MouseEvent event) {
				puzzle.requestFocus();
			}
		});
		setTitle(title);
		/* puzzle.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		puzzle.setSize(widthValue, heightValue);
	}

	public static void main(String[] args) {
		String laf = UIManager.getCrossPlatformLookAndFeelClassName();

		try {
			UIManager.setLookAndFeel(laf);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
			e.printStackTrace();
			System.exit(1);
		}
		List<Image> iconList = new ArrayList<>();
		for (String icon : PyraminxInterface.ICONS) {
			URL url = PyraminxFrame.class.getResource("/" + icon);
			if (url != null)
				iconList.add(Toolkit.getDefaultToolkit().getImage(url));
		}

		argumentParser = new ArgumentParser(args);
		if (argumentParser.hasOption("?")) {
			System.out.println("usage:" + NL +
"\t[-windowWidth={int}] [-windowHeight={int}]" + NL +
"\t[-windowX={int}] [-windowY={int}]" + NL +
"\t[-faceColor{0|1|2|3}={int}]" + NL +
"\t[-borderColor={int}] [-fgColor={int}] [-bgColor={int}]" + NL +
"\t[-sound={bool}] [-control={bool}] [-{size={int}]" + NL +
"\t[-mode={int} | -both}]" + NL +
"\t[-orient={bool}] [-sticky={bool}] [-practice={bool}]" + NL +
"\t[-delay={int}]");
			return;
		}
		PyraminxFrame frame = new PyraminxFrame(PyraminxInterface.TITLE);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		if (!iconList.isEmpty()) {
			frame.setIconImages(iconList);
		}
		frame.pack();
		frame.puzzle.requestFocusInWindow();
		frame.setLocation(PyraminxFrame.xValue, PyraminxFrame.yValue);
		frame.setVisible(true);
	}
}
